package checkers.flow;

import checkers.basetype.BaseTypeChecker;
import checkers.source.SourceChecker;
import checkers.types.*;
import checkers.types.AnnotatedTypeMirror.*;
import checkers.util.*;

import java.io.PrintStream;
import java.util.*;

import com.sun.source.tree.*;
import com.sun.source.util.*;

import javax.annotation.processing.ProcessingEnvironment;
import javax.lang.model.element.*;
import javax.lang.model.type.*;
import javax.lang.model.util.Elements;

/**
 * Provides a generalized flow-sensitive qualifier inference for the checkers
 * framework.
 *
 * <p>
 *
 * This implementation is based largely on {@code javac}'s dataflow analysis
 * module, which may be found in {@code com.sun.tools.javac.comp.Flow}. Note
 * that, whenever possible, this implementation prefers the use of classes in
 * the public API ({@link BitSet}, the Compiler Tree API) over those in
 * {@code com.sun.tools.javac}.
 *
 * <p>
 *
 * A separate instance of the analysis must be created for each compilation
 * unit.
 */
public class Flow extends TreePathScanner<Void, Void> {

    /** Where to print debugging messages, if non-null. */
    public static PrintStream debug = null;
    
    /** The checker to which this instance belongs. */
    protected final SourceChecker checker;

    /** The processing environment to use. */
    protected final ProcessingEnvironment env;

    /** The file that's being analyzed. */
    protected final CompilationUnitTree root;

    /** The annotations to infer. */
    protected final Set<AnnotationMirror> annotations;

    /** Utility class for getting source positions. */
    protected final SourcePositions source;

    /** Utility class for determining annotated types. */
    protected final AnnotatedTypeFactory factory;
    
    /** Utility class for operations on annotated types. */
    protected final AnnotatedTypes atypes;

    /** Maps variables to a bit index. */
    protected final List<VariableElement> vars;

    /** Stores the results of the analysis (source position to status bit). */
    protected final Map<Location, AnnotationMirror> record;

    /** Tracks the annotated state of each variable during flow. */
    protected GenKillGroup<AnnotationMirror> anno;

    /** Tracks the annotated state of each variable in a true branch. */
    protected GenKillGroup<AnnotationMirror> annoWhenTrue;

    /** Tracks the annotated state of each variable in a false branch. */
    protected GenKillGroup<AnnotationMirror> annoWhenFalse;

    /** Minimal liveness bit, used for better tracking with jumps. */
    private boolean alive = true;

    /** Tracks annotations in try blocks to support exceptions. */
    private final Deque<GenKillGroup<AnnotationMirror>> tryBits;

    /** Visitor state; tracking is required for checking receiver types. */
    private final VisitorState visitorState;

    protected final AnnotationRelations annoRelations;

    /** Utilities for {@link Element}s. */
    protected final Elements elements;

    /**
     * Creates a new analysis. The analysis will use the given {@link
     * AnnotatedTypeFactory} to obtain annotated types.
     *
     * @param checker the current checker
     * @param root the compilation unit that will be scanned
     * @param annotations the annotations to track
     * @param factory the factory class that will be used to get annotated
     *        types, or {@code null} if the default factory should be used
     */
    public Flow(BaseTypeChecker checker, CompilationUnitTree root,
            Set<AnnotationMirror> annotations, AnnotatedTypeFactory factory) {

        this.checker = checker;
        this.env = checker.getProcessingEnvironment();
        this.root = root;
        this.annotations = annotations;

        this.source = Trees.instance(env).getSourcePositions();
        if (factory == null)
            this.factory = new AnnotatedTypeFactory(checker, root);
        else this.factory = factory;
        
        this.atypes = new AnnotatedTypes(env, factory);
        
        this.visitorState = factory.getVisitorState();

        this.vars = new ArrayList<VariableElement>();
        this.record = new HashMap<Location, AnnotationMirror>();

        this.anno = new GenKillGroup<AnnotationMirror>(this.annotations);
        this.annoWhenTrue = null;
        this.annoWhenFalse = null;
        
        this.tryBits = new LinkedList<GenKillGroup<AnnotationMirror>>();
        
        this.annoRelations = checker.getAnnotationRelations();
        elements = env.getElementUtils();
    }

    @Override
    public Void scan(Tree tree, Void p) {
        if (tree != null && getCurrentPath() != null)
            this.visitorState.setPath(new TreePath(getCurrentPath(), tree));
        return super.scan(tree, p);
    }

    /**
     * Tests a result for a particular tree.
     * 
     * @param tree the tree to test
     * @return {@link Boolean#TRUE} if the bit for that position is set, {@link
     * Boolean#FALSE} if the bit for that position is not set, or null if there
     * is no result for that position
     */
    public AnnotationMirror test(Tree tree) {
        long pos = source.getStartPosition(root, tree);
        final Location loc = new Location(pos, tree.getKind());
        if (!record.containsKey(loc))
            return null;
        return record.get(loc);
    }

    /**
     * Registers a new variable for flow tracking using its definition.
     *
     * @param tree the variable to register
     */
    void newVar(VariableTree tree) {

        VariableElement var = TreeUtils.elementFromDeclaration(tree);
        assert var != null : "no symbol from tree";

        if (vars.contains(var)) {
            if (debug != null)
                debug.println("Flow: newVar(" + tree + ") reusing index");
            return;
        }

        int idx = vars.size();
        vars.add(var);

        AnnotatedTypeMirror type = factory.getAnnotatedType(tree);
        assert type != null : "no type from symbol";

        if (debug != null)
            debug.println("Flow: newVar(" + tree + ") -- " + type);

        // Determine the initial status of the variable by checking its
        // annotated type.
        for (AnnotationMirror annotation : annotations) {
            if (hasAnnotation(type, annotation))
                anno.set(annotation, idx);
            else
                anno.clear(annotation, idx);
        }
    }
    
    /**
     * Determines whether a type has an annotation. If the type is not a
     * wildcard, it checks the type directly; if it is a wildcard, it checks the
     * wildcard's "extends" bound (if it has one).
     * 
     * @param type the type to check
     * @param annotation the annotation to check for
     * @return true if the (non-wildcard) type has the annotation or, if a
     *         wildcard, the type has the annotation on its extends bound
     */
    private boolean hasAnnotation(AnnotatedTypeMirror type, 
            AnnotationMirror annotation) {       
        if (!(type instanceof AnnotatedWildcardType))
            return type.hasAnnotation(annotation);
        AnnotatedWildcardType wc = (AnnotatedWildcardType) type;
        AnnotatedTypeMirror bound = wc.getExtendsBound();
        if (bound != null && bound.hasAnnotation(annotation))
            return true;
        return false;
    }

    /**
     * Moves bits as assignments are made.
     *
     * @param lhs the left-hand side of the assignment
     * @param rhs the right-hand side of the assignment
     */
    void propagate(Tree lhs, Tree rhs) {

        if (debug != null) 
            debug.println("Flow: try propagate from " + rhs);
        
	// Skip assignment to arrays.
	if (lhs.getKind() == Tree.Kind.ARRAY_ACCESS)
	    return;
	
        // Get the element for the left-hand side.
        Element elt = InternalUtils.symbol(lhs);
        assert elt != null;

        // Get the element for the right hand side.
        AnnotatedTypeMirror type = factory.getAnnotatedType(rhs);
        assert type != null;

        int idx = vars.indexOf(elt);
        assert idx >= 0;
        if (idx < 0) return;

        Element rElt = InternalUtils.symbol(rhs);/* TODO type.getElement();*/
        int rIdx = vars.indexOf(rElt);

//        if (debug != null)
//            debug.println("Flow: (tree) propagate to " + lhs 
//                + " from " + type + (rIdx >= 0 ? "/" + anno.get(rIdx) : ""));

        for (AnnotationMirror annotation : annotations) {
            // Propagate/clear the bit if it's annotated or had its bit set.
            if (hasAnnotation(type, annotation) || (rIdx >= 0 && anno.get(annotation, rIdx)))
                anno.set(annotation, idx);
            else anno.clear(annotation, idx);
        }
    }
    
    /**
     * Moves bits in an assignment using a type instead of a tree.
     * 
     * @param lhs the left-hand side of the assignment
     * @param rhs the type of the right-hand side of the assignment
     */
    void propagateFromType(Tree lhs, AnnotatedTypeMirror rhs) {
        
        if (lhs.getKind() == Tree.Kind.ARRAY_ACCESS)
            return;
        
        Element elt = InternalUtils.symbol(lhs);
        
        int idx = vars.indexOf(elt);
        if (idx < 0) return;
        
//        if (debug != null)
//            debug.println("Flow: (type) propagate to " + lhs 
//                + " from " + rhs + " (" + rhs.hasAnnotation(this.annotation) + ") "
//                + (idx >= 0 ? "/" + anno.get(idx) : ""));
        
        for (AnnotationMirror annotation : annotations) {
            if (hasAnnotation(rhs, annotation))
                anno.set(annotation, idx);
            else anno.clear(annotation, idx);
        }
    }

    /**
     * Split the bitset before a conditional branch.
     */
    void split() {
        annoWhenFalse = GenKillGroup.copy(anno);
        annoWhenTrue = anno;
        anno = null;
    }

    /**
     * Merge the bitset after a conditional branch.
     */
    void merge() {
        anno = GenKillGroup.copy(anno);
        anno.and(annoWhenFalse);
        //annoWhenTrue = annoWhenFalse = null;
    }

    /**
     * @param path the path to check
     * @return true if the path leaf is part of an expression used as an lvalue
     */
    private boolean isLValue(TreePath path) {
        Tree last = null;
        for (Tree tree : path) {
            if (tree instanceof AssignmentTree)
                return last == ((AssignmentTree)tree).getVariable();
            else if (tree instanceof CompoundAssignmentTree)
                return last == ((CompoundAssignmentTree)tree).getVariable();
            if (last != null) break;
            last = tree;
        }
        return false;
    }
    
    /**
     * Record the value of the annotation bit for the given usage of a
     * variable, so that a type-checker may use its value after the analysis
     * has finished.
     *
     * @param path
     */
    protected void recordBits(TreePath path) {

        if (isLValue(path))
            return;
        
        Tree tree = path.getLeaf();

        Element elt = null;
        if (tree instanceof MemberSelectTree)
            elt = TreeUtils.elementFromUse((MemberSelectTree)tree);
        else if (tree instanceof IdentifierTree)
            elt = TreeUtils.elementFromUse((IdentifierTree)tree);
        else return;

        int idx = vars.indexOf(elt);
        if (idx < 0 && elt instanceof VariableElement) {
            idx = vars.size();
            vars.add((VariableElement)elt);
        }

        if (idx >= 0) {
            long pos = source.getStartPosition(root, tree);
            for (AnnotationMirror annotation : annotations) { 
                if (debug != null)
                    debug.println("Flow: recordBits(" + tree + ") @" + pos + " " + annotation + " "
                            + anno.get(annotation, idx) + " as " + tree.getKind());
                if (anno.get(annotation, idx)) {
                    Location loc = new Location(pos, tree.getKind());
                    AnnotationMirror existing = record.get(loc);
                    
                    // Don't replace the existing annotation unless the current
                    // annotation is *more* specific than the existing one.
                    if (existing == null || annoRelations.isSubtype(annotation, existing))
                        record.put(loc, annotation);
                }
            }
        }
    }

    /**
     * Called whenever a definition is scanned.
     *
     * @param tree the definition being scanned
     */
    protected void scanDef(Tree tree) {
        alive = true;
        scan(tree, null);
    }

    /**
     * Called whenever a statement is scanned.
     *
     * @param tree the statement being scanned
     */
    protected void scanStat(StatementTree tree) {
        alive = true;
        scan(tree, null);
    }

    /**
     * Called whenever a block of statements is scanned.
     *
     * @param trees the statements being scanned
     */
    protected void scanStats(List<? extends StatementTree> trees) {
        scan(trees, null);
    }

    /**
     * Called whenever a conditional expression is scanned.
     *
     * @param tree the condition being scanned
     */
    protected void scanCond(Tree tree) {
        alive = true;
        scan(tree, null);
        if (anno != null) split();
        anno = null;
    }

    /**
     * Called whenever an expression is scanned.
     *
     * @param tree the expression being scanned
     */
    protected void scanExpr(ExpressionTree tree) {
        alive = true;
        scan(tree, null);
        if (anno == null) merge();
    }

    /**
     * Called whenever a group of expressions is scanned.
     *
     * @param trees the expressions being scanned
     */
    protected void scanExprs(List<? extends ExpressionTree> trees) {
        for (ExpressionTree tree : trees)
            scanExpr(tree);

    }

    @Override
    public Void visitClass(ClassTree node, Void p) {
        AnnotatedDeclaredType preACT = visitorState.getClassType();
        ClassTree preCT = visitorState.getClassTree();

        visitorState.setClassType(factory.getAnnotatedType(node));
        visitorState.setClassTree(node);
        
        // Ensure that all fields are scanned before scanning methods.
        try { 
            scan(node.getModifiers(), p);
            scan(node.getTypeParameters(), p);
            scan(node.getExtendsClause(), p);
            scan(node.getImplementsClause(), p);
            for (Tree t : node.getMembers()) {
                if (t.getKind() == Tree.Kind.METHOD) continue;
                scan(t, p);
            }
            for (Tree t : node.getMembers()) {
                if (t.getKind() != Tree.Kind.METHOD) continue;
                scan(t, p);
            }
            return null;
        } finally {
            this.visitorState.setClassType(preACT);
            this.visitorState.setClassTree(preCT);
        }
    }

    @Override
    public Void visitImport(ImportTree tree, Void p) {
        return null;
    }

    @Override
    public Void visitInstanceOf(InstanceOfTree tree, Void p) {
        super.visitInstanceOf(tree, p);

        ExpressionTree expr = tree.getExpression();

        Element elt = null;
        if (expr instanceof IdentifierTree)
            elt = TreeUtils.elementFromUse((IdentifierTree) expr);
        else if (expr instanceof MemberSelectTree)
            elt = TreeUtils.elementFromUse((MemberSelectTree) expr);

        if (elt != null && vars.contains(elt)) {
            int idx = vars.indexOf(elt);
            for (AnnotationMirror annotation : annotations)
                if (hasAnnotation(factory.getAnnotatedTypeFromTypeTree(tree.getType()), annotation))
                    anno.set(annotation, idx);
        }

        return null;
    }

    @Override
    public Void visitTypeCast(TypeCastTree node, Void p) {
        super.visitTypeCast(node, p);
        if (!factory.fromTypeTree(node.getType()).getAnnotations().isEmpty())
            return null;
        AnnotatedTypeMirror t = factory.getAnnotatedType(node.getExpression());
        long pos = source.getStartPosition(root, node);
        for (AnnotationMirror a : annotations)
            if (hasAnnotation(t, a))
                record.put(new Location(pos, node.getKind()), a);
        return null;
    }

    @Override
    public Void visitAnnotation(AnnotationTree tree, Void p) {
        return null;
    }

    @Override
    public Void visitIdentifier(IdentifierTree node, Void p) {
        super.visitIdentifier(node, p);
        recordBits(getCurrentPath());
        return null;
    }

    @Override
    public Void visitMemberSelect(MemberSelectTree node, Void p) {
        super.visitMemberSelect(node, p);
        recordBits(getCurrentPath());
        return null;
    }

    @Override
    public Void visitVariable(VariableTree node, Void p) {
        newVar(node);
        ExpressionTree init = node.getInitializer();
        if (init != null) {
            VariableElement elem = TreeUtils.elementFromDeclaration(node);
            if (!isNonFinalField(elem)) {
                scanExpr(init);
                propagate(node, init);
            }
        }
        return null;
    }

    /**
     * Tests whether the element is of a non-final field
     * 
     * @return true iff element is a non-final field
     */
    private static final boolean isNonFinalField(Element element) {
        return (element.getKind().isField()
                && !ElementUtils.isFinal(element));
    }

    @Override
    public Void visitAssignment(AssignmentTree node, Void p) {
        ExpressionTree var = node.getVariable();
        ExpressionTree expr = node.getExpression();
        if (!(var instanceof IdentifierTree))
            scanExpr(var);
        scanExpr(expr);
        propagate(var, expr);
        return null;
    }
    
    // This is an exact copy of visitAssignment()
    @Override
    public Void visitCompoundAssignment(CompoundAssignmentTree node, Void p) {
        ExpressionTree var = node.getVariable();
        ExpressionTree expr = node.getExpression();
        if (!(var instanceof IdentifierTree))
            scanExpr(var);
        scanExpr(expr);
        propagate(var, expr);
        return null;
    }

    @Override
    public Void visitEnhancedForLoop(EnhancedForLoopTree node, Void p) {
        VariableTree var = node.getVariable();
        newVar(var);
        
        ExpressionTree expr = node.getExpression();
        scanExpr(expr);
        
        AnnotatedTypeMirror rhs = factory.getAnnotatedType(expr);
        AnnotatedTypeMirror iter = atypes.getIteratedType(rhs);
        if (iter != null)
            propagateFromType(var, iter);
        
        return super.visitEnhancedForLoop(node, p);
    }

    @Override
    public Void visitAssert(AssertTree node, Void p) {
        scanCond(node.getCondition());
        anno = GenKillGroup.copy(annoWhenTrue);
        return null;
    }

    @Override
    public Void visitIf(IfTree node, Void p) {
        scanCond(node.getCondition());

        GenKillGroup<AnnotationMirror> before = annoWhenFalse;
        anno = annoWhenTrue;

        boolean aliveBefore = alive;

        scanStat(node.getThenStatement());
        StatementTree elseStmt = node.getElseStatement();
        if (elseStmt != null) {
            boolean aliveAfter = alive;
            alive = aliveBefore;
            GenKillGroup<AnnotationMirror> after = GenKillGroup.copy(anno);
            anno = before;
            scanStat(elseStmt);
            alive &= aliveAfter;
            if (!alive)
                anno = GenKillGroup.copy(after);
            else
                anno.and(after);
        } else {
            alive &= aliveBefore;
            if (!alive)
                anno = GenKillGroup.copy(before);
            else
                anno.and(before);
        }

        return null;
    }
    
    @Override
    public Void visitConditionalExpression(ConditionalExpressionTree node,
            Void p) {
        
        // Split and merge as for an if/else.
        scanCond(node.getCondition());
        
        GenKillGroup<AnnotationMirror> before = annoWhenFalse;
        anno = annoWhenTrue;

        scanExpr(node.getTrueExpression());
        GenKillGroup<AnnotationMirror> after = GenKillGroup.copy(anno);
        anno = before;
        
        scanExpr(node.getFalseExpression());
        anno.and(after);
        
        return null;
    }

    @Override
    public Void visitWhileLoop(WhileLoopTree node, Void p) {
        boolean pass = false;
        GenKillGroup<AnnotationMirror> annoCond;
        do {
            GenKillGroup<AnnotationMirror> annoEntry = GenKillGroup.copy(anno);
            scanCond(node.getCondition());
            annoCond = annoWhenFalse;
            anno = annoWhenTrue;
            scanStat(node.getStatement());
            if (pass) break;
            annoWhenTrue.and(annoEntry);
            pass = true;
        } while (true);
        anno = annoCond;
        return null;
    }
    
    @Override
    public Void visitBreak(BreakTree node, Void p) {
        alive = false;
        return null;
    }

    @Override
    public Void visitContinue(ContinueTree node, Void p) {
        alive = false;
        return null;
    }

    @Override
    public Void visitReturn(ReturnTree node, Void p) {
        if (node.getExpression() != null)
            scanExpr(node.getExpression());
        alive = false;
        return null;
    }

    @Override
    public Void visitThrow(ThrowTree node, Void p) {
        scanExpr(node.getExpression());
        alive = false;
        return null;
    }

    @Override
    public Void visitTry(TryTree node, Void p) {
	tryBits.push(GenKillGroup.copy(anno));
	scan(node.getBlock(), p);
	GenKillGroup<AnnotationMirror> annoAfterBlock = GenKillGroup.copy(anno);
	GenKillGroup<AnnotationMirror> result = tryBits.pop();
	anno.and(result);
	if (node.getCatches() != null) {
	    boolean catchAlive = true;
	    for (CatchTree ct : node.getCatches()) {
	        scan(ct, p);
	        catchAlive &= alive;
	    }
	    // Conservative: only if there's no finally
	    if (!catchAlive && node.getFinallyBlock() == null)
	        anno = GenKillGroup.copy(annoAfterBlock);
	}
	scan(node.getFinallyBlock(), p);
	return null;
    }

    @Override
    public Void visitMethodInvocation(MethodInvocationTree node, Void p) {
        super.visitMethodInvocation(node, p);

        ExecutableElement method = TreeUtils.elementFromUse(node);
        if (method.getSimpleName().contentEquals("exit")
                && method.getEnclosingElement().getSimpleName().contentEquals("System"))
            alive = false;
        
        final String methodPackage = elements.getPackageOf(method).getQualifiedName().toString();
        boolean isJDKMethod = methodPackage.startsWith("java") || methodPackage.startsWith("com.sun");
        for (int i = 0; i < vars.size(); i++) {
            Element var = vars.get(i);
            for (AnnotationMirror a : annotations)
                if (!isJDKMethod && isNonFinalField(var) && !varDefHasAnnotation(a, var))
                    anno.clear(a, i);
        }
        

	List<? extends TypeMirror> thrown = method.getThrownTypes();
	if (!thrown.isEmpty() 
	        && TreeUtils.enclosingOfKind(getCurrentPath(), Tree.Kind.TRY) != null) {
	    if (!tryBits.isEmpty())
	        tryBits.peek().and(anno);
	}

        return null;
    }
    
    private Map<Element, Boolean> annotatedVarDefs = new HashMap<Element, Boolean>();
   
    /** 
     * Determines whether a variable definition has been annotated.
     * 
     * @param annotation the annotation to check for
     * @param var the variable to check
     * @return true if the variable has the given annotation, false otherwise
     */
    private boolean varDefHasAnnotation(AnnotationMirror annotation, Element var) {
        
        if (annotatedVarDefs.containsKey(var))
            return annotatedVarDefs.get(var);
        
        boolean result = hasAnnotation(factory.getAnnotatedType(var), annotation);
        annotatedVarDefs.put(var, result);
        return result;
    }

    @Override
    public Void visitBlock(BlockTree node, Void p) {
        if (node.isStatic()) {
            GenKillGroup<AnnotationMirror> prev = GenKillGroup.copy(anno);
            try {
                super.visitBlock(node, p);
                return null;
            } finally {
                anno = prev;
            }
        }
        return super.visitBlock(node, p);
    }

    @Override
    public Void visitMethod(MethodTree node, Void p) {
        AnnotatedDeclaredType preMRT = visitorState.getMethodReceiver();
        MethodTree preMT = visitorState.getMethodTree();
        visitorState.setMethodReceiver(
                factory.getAnnotatedType(node).getReceiverType());
        visitorState.setMethodTree(node);
        
        // Intraprocedural, so save and restore bits.
        GenKillGroup<AnnotationMirror> prev = GenKillGroup.copy(anno);
        try {
            super.visitMethod(node, p);
            return null;
        } finally {
            anno = prev;
            visitorState.setMethodReceiver(preMRT);
            visitorState.setMethodTree(preMT);
        }
    }
}
