package checkers.javari;

import java.util.*;

import javax.annotation.processing.ProcessingEnvironment;
import javax.lang.model.element.*;
import javax.lang.model.type.TypeKind;

import checkers.types.*;
import checkers.util.TreeUtils;
import checkers.util.AnnotationUtils;
import static checkers.types.AnnotatedTypeMirror.*;

import com.sun.source.tree.*;
import com.sun.source.util.SimpleTreeVisitor;
import com.sun.tools.javac.tree.JCTree;

import checkers.quals.*;

/**
 * Generates AnnotatedTypeMirrors with Javari annotations from Trees and Elements.
 */
public class JavariAnnotatedTypeFactory extends AnnotatedTypeFactory {

    /** Adds annotations from tree context before type resolution. */
    private final TreePreAnnotator treePre;

    /** Adds annotations from the resulting type after type resolution. */
    private final TypePostAnnotator typePost;

    /** The Javari annotations. */
    private final AnnotationMirror READONLY, THISMUTABLE, MUTABLE, POLYREAD, QREADONLY, ASSIGNABLE;

    /**
     * Creates a new {@link JavariAnnotatedTypeFactory} that operates on a
     * particular AST.
     *
     * @param checker the checker to which this factory belongs
     * @param root the AST on which this type factory operates
     */
    public JavariAnnotatedTypeFactory(JavariChecker checker,
        CompilationUnitTree root) {
        super(checker, root);
        this.treePre = new TreePreAnnotator();
        this.typePost = new TypePostAnnotator();
        this.READONLY = annotations.fromName(ReadOnly.class.getCanonicalName());
        this.THISMUTABLE = annotations.fromName(ThisMutable.class.getCanonicalName());
        this.MUTABLE = annotations.fromName(Mutable.class.getCanonicalName());
        this.POLYREAD = annotations.fromName(PolyRead.class.getCanonicalName());
        this.QREADONLY = annotations.fromName(QReadOnly.class.getCanonicalName());
        this.ASSIGNABLE = annotations.fromName(Assignable.class.getCanonicalName());
    }

    /**
     * Returns the annotation specifying the immutability type of {@code type}.
     */
    private AnnotationMirror getImmutabilityAnnotation(/*@ReadOnly*/ AnnotatedTypeMirror type) {
        if (type.hasAnnotation(READONLY))
            return READONLY;
        else if (type.hasAnnotation(MUTABLE))
            return MUTABLE;
        else if (type.hasAnnotation(POLYREAD))
            return POLYREAD;
        else if (type.hasAnnotation(QREADONLY))
            return QREADONLY;
        else if (type.hasAnnotation(THISMUTABLE))
            return THISMUTABLE;
        else
            return null;
    }

    /**
     * @param type  an annotated type mirror
     * @return  true iff the type is specified an immutability type other than this-mutable,
     *          false otherwise
     */
    public boolean hasImmutabilityAnnotation(/*@ReadOnly*/ AnnotatedTypeMirror type) {
        return type != null && getImmutabilityAnnotation(type) != null;
    }


    @Override
    protected void annotateImplicit(Tree tree, /*@Mutable*/ AnnotatedTypeMirror type) {
        treePre.visit(tree, type);
        if (tree.getKind() == Tree.Kind.CLASS) {
            if (!hasImmutabilityAnnotation(type))
                type.addAnnotation(MUTABLE);
        } else
            typePost.visit(type);

        if (type.hasAnnotation(THISMUTABLE)) {

            AnnotatedDeclaredType mReceiver = getCurrentMethodReceiver(tree);
            if (mReceiver != null) {
                if (mReceiver.hasAnnotation(POLYREAD)) {
                    new SimpleAnnotatedTypeScanner<Void, Void>() {
                        @Override
                            protected Void defaultAction(AnnotatedTypeMirror type, Void p) {
                            if (type.hasAnnotation(THISMUTABLE)) {
                                type.removeAnnotation(THISMUTABLE);
                                type.addAnnotation(POLYREAD);
                            }
                            return super.defaultAction(type, p);
                        }
                    }.visit(type);
                }
                else if (!mReceiver.hasAnnotation(READONLY)) {
                    new SimpleAnnotatedTypeScanner<Void, Void>() {
                        @Override
                            protected Void defaultAction(AnnotatedTypeMirror type, Void p) {
                            if (type.hasAnnotation(THISMUTABLE)) {
                                type.removeAnnotation(THISMUTABLE);
                                type.addAnnotation(MUTABLE);
                            }
                            return super.defaultAction(type, p);
                        }
                    }.visit(type);
                }
            }
        }
    }

    @Override
    protected void annotateImplicit(Element element, AnnotatedTypeMirror type) {
        if (element.getKind().isClass() || element.getKind().isInterface()) {
            if (!hasImmutabilityAnnotation(type))
                type.addAnnotation(MUTABLE);
        } else
            typePost.visit(type);
    }

    @Override
    protected void annotateInheritedFromClass(/*@Mutable*/ AnnotatedTypeMirror type) {
        // Do nothing
    }


    /**
     * Returns a type of the class and the field {@code this}.
     */
    @Override
    public AnnotatedDeclaredType getSelfType(Tree tree) {
        AnnotatedDeclaredType act = this.getCurrentClassType(tree);
        AnnotatedDeclaredType methodReceiver = this.getCurrentMethodReceiver(tree);

        if (methodReceiver == null || !hasImmutabilityAnnotation(methodReceiver))
            return act;

        act.removeAnnotations(act.getAnnotations());
        act.addAnnotations(methodReceiver.getAnnotations());

        return act;
    }


    /**
     * Returns a singleton collection with the most restrictive immutability
     * annotation that is a supertype of the annotations on both collections.
     */
    @Override
    public Collection<AnnotationMirror> unify(Collection<AnnotationMirror> c1,
            Collection<AnnotationMirror> c2) {
        Map<String, AnnotationMirror> ann =
            new HashMap<String, AnnotationMirror>();
        for (AnnotationMirror anno : c1)
            ann.put(AnnotationUtils.annotationName(anno).toString(), anno);
        for (AnnotationMirror anno : c2)
            ann.put(AnnotationUtils.annotationName(anno).toString(), anno);

        if (ann.containsKey(QReadOnly.class.getCanonicalName()))
            return Collections.singleton(QREADONLY);
        else if (ann.containsKey(ReadOnly.class.getCanonicalName()))
            return Collections.singleton(READONLY);
        else if (ann.containsKey(PolyRead.class.getCanonicalName()))
            return Collections.singleton(POLYREAD);
        else
            return Collections.singleton(MUTABLE);
    }



    @Override
    public AnnotatedExecutableType methodFromUse(MethodInvocationTree tree) {
        AnnotatedExecutableType type = super.methodFromUse(tree);

        ExecutableElement executableElt = type.getElement();
        AnnotatedTypeMirror returnType = type.getReturnType();

        if (returnType.hasAnnotation(POLYREAD)) {

            List<AnnotatedTypeMirror> parameterTypes = type.getParameterTypes();
            List<? extends ExpressionTree> argumentTrees = tree.getArguments();
            AnnotatedTypeMirror receiverType = type.getReceiverType();

            boolean allMutable = true, allPolyRead = true, allThisMutable = true;

            // look at parameters and arguments
            // TODO: deal with varargs parameters
            for (int i = 0; i < parameterTypes.size(); i++) {
                AnnotatedTypeMirror pType = parameterTypes.get(i);
                // only look at it if parameter is PolyRead
                if (pType.hasAnnotation(POLYREAD)) {
                    AnnotatedTypeMirror aType = fromExpression(argumentTrees.get(i));

                    if (aType.hasAnnotation(THISMUTABLE))
                        allMutable = false;

                    if (aType.hasAnnotation(READONLY) || aType.hasAnnotation(QREADONLY)) {
                        allMutable = false; allThisMutable = false;
                    }

                    if (!(aType.hasAnnotation(POLYREAD)
                          && !aType.hasAnnotation(READONLY)
                          && !aType.hasAnnotation(MUTABLE)
                          && !aType.hasAnnotation(QREADONLY)))
                        allPolyRead = false;

                }
            }

            // look at receiver type and reference
            if (receiverType.hasAnnotation(POLYREAD)) {
                // if MemberSelectTree, we can just look at the expression tree
                ExpressionTree exprTree = tree.getMethodSelect();
                if (exprTree.getKind() == Tree.Kind.MEMBER_SELECT) {
                    MemberSelectTree mst = (MemberSelectTree) exprTree;
                    AnnotatedTypeMirror referenceType = getAnnotatedType(mst.getExpression());

                    if (referenceType.hasAnnotation(READONLY)) {
                        allMutable = false;
                        allThisMutable = false;
                        allPolyRead = false;
                    } else if (referenceType.hasAnnotation(POLYREAD)) {
                        allMutable = false;
                    }

                } else {
                    // not MemberSelectTree, get context from method's enclosing class
                    AnnotatedTypeMirror classType
                        = fromElement(executableElt.getEnclosingElement());
                    if (classType.hasAnnotation(READONLY)) {
                        allMutable = false;
                        allThisMutable = false;
                        allPolyRead = false;
                    } else if (classType.hasAnnotation(POLYREAD)) {
                        allMutable = false;
                    }
                }
            }

            // now, add appropriate type to return list
            returnType.removeAnnotation(POLYREAD);
            if (allMutable) returnType.addAnnotation(MUTABLE);
            else if (allPolyRead) returnType.addAnnotation(POLYREAD);
            else if (allThisMutable) returnType.addAnnotation(THISMUTABLE);
            else returnType.addAnnotation(READONLY);

        } else { // return type has no PolyRead, just get the annotated return type
            if (returnType.hasAnnotation(READONLY))returnType.addAnnotation(READONLY);
            else if (!returnType.getKind().isPrimitive()) returnType.addAnnotation(MUTABLE);
            else returnType.addAnnotation(THISMUTABLE);
        }

        Map<AnnotatedTypeMirror, AnnotatedTypeMirror> returnTypeMap
            = Collections.singletonMap(type.getReturnType(), returnType);
        type = type.substitute(returnTypeMap);

        return type;
    }

    @Override
    public void postAsMemberOf(AnnotatedTypeMirror type,
            AnnotatedTypeMirror owner, Element element) {
        if (!owner.hasAnnotation(READONLY)) {
            final AnnotationMirror ownerAnno = getImmutabilityAnnotation(owner);
            new SimpleAnnotatedTypeScanner<Void, Void>() {
                @Override
                protected Void defaultAction(AnnotatedTypeMirror type, Void p) {
                    if (type.hasAnnotation(THISMUTABLE)) {
                        type.removeAnnotation(THISMUTABLE);
                        type.addAnnotation(ownerAnno);
                    }
                    return super.defaultAction(type, p);
                }
            }.visit(type);
        }
    }


    /**
     * A visitor to get the annotations from a tree.
     */
    private class TreePreAnnotator extends SimpleTreeVisitor<Void, AnnotatedTypeMirror> {

        /** Default action: do nothing */
        @Override
        public Void defaultAction(Tree node, AnnotatedTypeMirror p) {
            return null;
        }


        /** Ensures every literal type has a {code @Mutable} annotation. */
        @Override
        public Void visitLiteral(LiteralTree node, AnnotatedTypeMirror p) {
            p.addAnnotation(MUTABLE);
            return super.visitLiteral(node, p);
        }


        /** Ensures the result has a {code @Mutable} annotation. */
        @Override
        public Void visitBinary(BinaryTree node, AnnotatedTypeMirror p) {
            p.addAnnotation(MUTABLE);
            return super.visitBinary(node, p);
        }

        @Override
        public Void visitNewArray(NewArrayTree node, AnnotatedTypeMirror p) {
            p.addAnnotation(MUTABLE);
            return super.visitNewArray(node, p);
        }

        /** Ensures the result has a {code @Mutable} annotation. */
        @Override
        public Void visitUnary(UnaryTree node, AnnotatedTypeMirror p) {
            p.addAnnotation(MUTABLE);
            return super.visitUnary(node, p);
        }

        /** Ensures the result has the internal this-mutable annotation, if nothing else */
        @Override
        public Void visitPrimitiveType(PrimitiveTypeTree node, AnnotatedTypeMirror p) {
            if (!hasImmutabilityAnnotation(p))
                p.addAnnotation(THISMUTABLE);
            return super.visitPrimitiveType(node, p);
        }

        @Override
        public Void visitIdentifier(IdentifierTree node, AnnotatedTypeMirror p) {
            if (!hasImmutabilityAnnotation(p)) {
                AnnotatedDeclaredType x = getCurrentMethodReceiver(node);
                if (x != null) {
                    if (x.hasAnnotation(MUTABLE)) p.addAnnotation(MUTABLE);
                    else if (x.hasAnnotation(POLYREAD)) p.addAnnotation(POLYREAD);
                    else p.addAnnotation(THISMUTABLE);
                } else p.addAnnotation(THISMUTABLE);
            }
            return super.visitIdentifier(node, p);
        }


        /** Returns list of annotations on the new class constructor; resolves {code @PolyRead} if applicable. */
        @Override
        public Void visitNewClass(NewClassTree node, AnnotatedTypeMirror p) {

            Element constructorElt = ((JCTree.JCNewClass)node).constructor;
            AnnotatedExecutableType executableType =
                (AnnotatedExecutableType) fromElement(constructorElt);
            AnnotatedTypeMirror receiverType = executableType.getReceiverType();

            // if receiver type has PolyRead, we must resolve it by looking at the arguments
            if (receiverType.hasAnnotation(POLYREAD)) {
                List<AnnotatedTypeMirror> parameterTypes = executableType.getParameterTypes();
                List<? extends ExpressionTree> argumentTrees = node.getArguments();

                boolean allMutable = true, allPolyRead = true, allThisMutable = true;

                for (int i = 0; i < parameterTypes.size(); i++) {
                    AnnotatedTypeMirror pType = parameterTypes.get(i);
                    if (pType.hasAnnotation(POLYREAD)) {
                        AnnotatedTypeMirror aType = fromExpression(argumentTrees.get(i));

                        if (aType.hasAnnotation(THISMUTABLE))
                            allMutable = false;

                        if (aType.hasAnnotation(READONLY) || aType.hasAnnotation(QREADONLY)) {
                            allMutable = false; allThisMutable = false;
                        }

                        if (!(aType.hasAnnotation(POLYREAD)
                              && !aType.hasAnnotation(READONLY)
                              && !aType.hasAnnotation(MUTABLE)
                              && !aType.hasAnnotation(QREADONLY)))
                            allPolyRead = false;
                    }
                }

                if (allMutable) {
                    if (!p.hasAnnotation(MUTABLE)) p.addAnnotation(MUTABLE);
                } else if (allPolyRead) {
                    if (!p.hasAnnotation(POLYREAD)) p.addAnnotation(POLYREAD);
                } else if (allThisMutable) {
                    if (!p.hasAnnotation(THISMUTABLE)) p.addAnnotation(THISMUTABLE);
                } else {
                    if (!p.hasAnnotation(READONLY)) p.addAnnotation(READONLY);
                }

            } else {   // if receiver has no polyread, just return the receiver type
                if (receiverType.hasAnnotation(READONLY)) p.addAnnotation(READONLY);
                if (receiverType.hasAnnotation(MUTABLE)) p.addAnnotation(MUTABLE);
            }

            return super.visitNewClass(node, p);
        }

        /**
         * Selects the appropriate annotation from the inner type, or inherits
         * it from the outer type according to the Javari typesystem.
         */
        @Override
        public Void visitMemberSelect(MemberSelectTree node, AnnotatedTypeMirror p) {

            AnnotatedTypeMirror outerType = getAnnotatedType(node.getExpression());
            AnnotatedTypeMirror innerType = fromElement(TreeUtils.elementFromUse(node));

            p.removeAnnotations(p.getAnnotations());
            if (innerType.hasAnnotation(READONLY))
                p.addAnnotation(READONLY);
            else if (innerType.hasAnnotation(MUTABLE))
                p.addAnnotation(MUTABLE);
            else if (innerType.hasAnnotation(POLYREAD) && outerType.hasAnnotation(POLYREAD))
                p.addAnnotation(POLYREAD);
            else if (outerType.hasAnnotation(READONLY))
                p.addAnnotation(READONLY);
            else if (outerType.hasAnnotation(MUTABLE))
                p.addAnnotation(MUTABLE);
            else if (outerType.hasAnnotation(POLYREAD))
                p.addAnnotation(POLYREAD);
            else
                p.addAnnotation(THISMUTABLE);

            return super.visitMemberSelect(node, p);
        }

    }


    /**
     * Responsible for adding implicit (not written on the code) annotations.
     */
    private class TypePostAnnotator extends AnnotatedTypeScanner<Void, Void> {

        /**
         * Passes down the enclosing class annotation, if any; else try
         * to retrieve from annotation if available, from enclosing
         * class if field, from supertype if class or interface,
         * or mutable otherwise.
         */
        @Override
        public Void scan(AnnotatedTypeMirror type, Void p) {
            // we don't want to add annotations if there are annotations already
            if (type != null && !hasImmutabilityAnnotation(type) && type.getElement() != null) {
                ElementKind eltKind = type.getElement().getKind();

                    if (eltKind.isField()) {
                    // inherit annotations from enclosing class
                    AnnotatedTypeMirror enclosingType =
                        fromElement(type.getElement().getEnclosingElement());

                    if (enclosingType.hasAnnotation(READONLY))
                        type.addAnnotation(READONLY);
                    else {
                        new SimpleAnnotatedTypeScanner<Void, Void>() {
                            @Override
                            protected Void defaultAction(AnnotatedTypeMirror type, Void p) {
                                if (hasImmutabilityAnnotation(type))
                                    return super.defaultAction(type, p);
                                if (type.getKind() == TypeKind.DECLARED) {
                                    AnnotatedDeclaredType dt = (AnnotatedDeclaredType)type;
                                    TypeElement tElt = (TypeElement) dt.getUnderlyingType().asElement();
                                    AnnotatedTypeMirror tType = getAnnotatedType(tElt);
                                    if (tType.hasAnnotation(READONLY))
                                        type.addAnnotation(READONLY);
                                }
                                if (!hasImmutabilityAnnotation(type))
                                    type.addAnnotation(THISMUTABLE);
                                return super.defaultAction(type, p);
                            }
                        }.visit(type);
                    }
                }

            }
            return super.scan(type, p);
        }

        @Override
        public Void visitExecutable(AnnotatedExecutableType type, Void p) {
            AnnotatedDeclaredType receiver = type.getReceiverType();
            if (!hasImmutabilityAnnotation(receiver)) {
                AnnotatedDeclaredType owner = (AnnotatedDeclaredType)
                    getAnnotatedType(type.getElement().getEnclosingElement());
                assert hasImmutabilityAnnotation(owner);
                receiver.addAnnotation(getImmutabilityAnnotation(owner));
            }
            return super.visitExecutable(type, p);
        }

        @Override
        public Void visitDeclared(AnnotatedDeclaredType type, Void p) {
            if (!hasImmutabilityAnnotation(type)) {
                TypeElement tElt = (TypeElement) type.getUnderlyingType().asElement();
                AnnotatedTypeMirror tType = getAnnotatedType(tElt);
                if (tType.hasAnnotation(READONLY)) {
                    type.addAnnotation(READONLY);
                    type.removeAnnotation(MUTABLE);
                } else
                    type.addAnnotation(MUTABLE);
            }
            return super.visitDeclared(type, p);
        }
    }
}
