package checkers.javari;

import checkers.basetype.*;
import checkers.source.SuppressWarningsKey;
import checkers.types.*;
import checkers.types.AnnotatedTypeMirror.AnnotatedDeclaredType;
import checkers.util.GraphAnnotationRelations;
import checkers.metaquals.TypeQualifiers;
import checkers.quals.*;

import javax.annotation.processing.*;
import javax.lang.model.*;
import javax.lang.model.element.*;

import com.sun.source.tree.*;

/**
 * An annotation processor that checks a program's use of the Javari
 * type annotations ({@code @ReadOnly}, {@code @Mutable}, {@code @Assignable},
 * {@code @PolyRead} and {@code @QReadOnly}). For specific
 * instructions on the usage of each of these annotations, see the
 * Javari language specification.
 */
@SupportedAnnotationTypes({"*"})
@SupportedSourceVersion(SourceVersion.RELEASE_7)
@SuppressWarningsKey("javari")
@TypeQualifiers( { ReadOnly.class, ThisMutable.class, Mutable.class,
    PolyRead.class, QReadOnly.class })
public class JavariChecker extends BaseTypeChecker {

    protected AnnotationMirror READONLY, THISMUTABLE, MUTABLE, POLYREAD, QREADONLY, ASSIGNABLE;

    /**
     * Initializes the checker: calls init method on super class,
     * creates a local AnnotationFactory based on the processing
     * environment, and uses it to create the protected
     * AnnotationMirrors used through this checker.
     * @param processingEnv the processing environment to use in the local AnnotationFactory
     */
    @Override
    public synchronized void init(ProcessingEnvironment processingEnv) {
        super.init(processingEnv);
        AnnotationFactory annoFactory = new AnnotationFactory(this.env);
        this.READONLY = annoFactory.fromName(ReadOnly.class.getCanonicalName());
        this.THISMUTABLE = annoFactory.fromName(ThisMutable.class.getCanonicalName());
        this.MUTABLE = annoFactory.fromName(Mutable.class.getCanonicalName());
        this.POLYREAD = annoFactory.fromName(PolyRead.class.getCanonicalName());
        this.QREADONLY = annoFactory.fromName(QReadOnly.class.getCanonicalName());
        this.ASSIGNABLE = annoFactory.fromName(Assignable.class.getCanonicalName());

        // This should go some where else!
        GraphAnnotationRelations annoRelations = new GraphAnnotationRelations(READONLY);
        annoRelations.addSubtype(MUTABLE, THISMUTABLE);
        annoRelations.addSubtype(MUTABLE, null);
        annoRelations.addSubtype(null, MUTABLE);    // makes null = Mutable
        annoRelations.addSubtype(READONLY, QREADONLY);
        annoRelations.addSubtype(MUTABLE, POLYREAD);
        annoRelations.addSubtype(null, POLYREAD);
        this.annoRelations = annoRelations;
    }

    @Override
    protected TypeRelations createTypeRelations() {
        return new TypeRelations(annoRelations) {
            @Override
            protected boolean compareTypeArg(AnnotatedTypeMirror lhs, AnnotatedTypeMirror rhs) {
                return lhs.hasAnnotation(QREADONLY) || super.compareTypeArg(lhs, rhs);
            }
         };
    }

    /** Return true if a type is primitive or if a type is subtype of another */
    @Override
    public boolean isSubtype(AnnotatedTypeMirror sup, AnnotatedTypeMirror sub) {
        return sub.getKind().isPrimitive() || super.isSubtype(sup, sub);
    }


    @Override
    public boolean isValidUse(AnnotatedTypeMirror elemType, AnnotatedTypeMirror useType) {
        return true;
    }

    /**
     * Returns a new JavariAnnotatedTypeFactory.
     *
     * @param env the ProcessingEnvironment to use in the factory initialization.
     * @param root the CompilationTreeUnit to use in the factory initialization.
     * @return a new JavariAnnotatedTypeFactory.
     */
    @Override
    public JavariAnnotatedTypeFactory getFactory(CompilationUnitTree root) {
        return new JavariAnnotatedTypeFactory(this, root);
    }
}
