package checkers.nonnull;

import javax.lang.model.element.*;

import checkers.basetype.*;
import checkers.source.Result;
import checkers.types.AnnotatedTypeMirror;
import checkers.types.AnnotatedTypeMirror.*;
import checkers.util.TreeUtils;

import com.sun.source.tree.*;

/**
 * A type-checking visitor for the {@link checkers.quals.NonNull} type
 * qualifier that uses the {@link BaseTypeVisitor} implementation. This visitor
 * reports errors or warnings for violations for the following cases:
 *
 * <ul>
 * <li>if the receiver of a member dereference is not NonNull (error
 * "dereference.of.nullable")</li>
 * </ul>
 * 
 * Self-accesses (via {@code this} or {@code super}) can always be dereferenced.
 */
public class NonNullVisitor extends BaseTypeVisitor<Void, Void> {

    private final AnnotationMirror NONNULL;
    private final AnnotatedTypeMirror NN_OBJECT;

    /**
     * Creates a new visitor for type-checking {@link checkers.quals.NonNull}.
     *
     * @param checker the checker to use
     * @param root the root of the input program's AST to check
     */
    public NonNullVisitor(NonNullChecker checker, CompilationUnitTree root) {
        super(checker, root);
        NONNULL = this.annoFactory.fromName("checkers.quals.NonNull");
        NN_OBJECT = factory.getAnnotatedType(elements.getTypeElement("java.lang.Object"));
        NN_OBJECT.clearAnnotations();
        NN_OBJECT.addAnnotation(NONNULL);
    }

    @Override
    public Void visitMemberSelect(MemberSelectTree node, Void p) {
    
        AnnotatedTypeMirror type = factory.getAnnotatedType(node.getExpression());
            
        if (!TreeUtils.isSelfAccess(node) && !checker.isSubtype(NN_OBJECT, type))
            checker.report(Result.failure("dereference.of.nullable",
                    node.getExpression().toString()), node);
                    
        return super.visitMemberSelect(node, p);
    }

    @Override
    public Void visitEnhancedForLoop(EnhancedForLoopTree node, Void p) {
        AnnotatedTypeMirror iterableType = 
            factory.getAnnotatedType(node.getExpression());
        if (!checker.isSubtype(NN_OBJECT, iterableType))
            checker.report(Result.failure("dereference.of.nullable", 
                    node.getExpression()), node.getExpression());
        return super.visitEnhancedForLoop(node, p);
    }

    @Override
    protected boolean checkMethodInvocability(AnnotatedExecutableType method,
            MethodInvocationTree node) {
        // Claim that methods with a @NonNull receiver are invocable so that
        // visitMemberSelect issues dereference errors instead.
        if (!TreeUtils.isSelfAccess(node) 
                && method.getReceiverType().hasAnnotation(NONNULL))
            return true;
        return super.checkMethodInvocability(method, node);
    }
}
